#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <compat/ina90.h>


#include <avr/delay.h>

#include <avr/io.h>
#include <compat/twi.h>		/* Note [1] */


/*
#define DI   0
#define DO   1
#define CLK  2 
#define CS   3
*/
#define DI   3
#define DO   0
#define CLK  2 
#define CS   1
//#define INS  4
//#define WP   5

#define DATA_BLOCK_SIZE			64
#define SHIFTER_DATA_BLOCK		6

#define MMC_BLOCK_SIZE	512

char buffer [DATA_BLOCK_SIZE]={};			//Buffer


// *do not change unless you know what you're doing
// Value may be SIGNAL or INTERRUPT
#ifndef UART_INTERRUPT_HANDLER
#define UART_INTERRUPT_HANDLER	SIGNAL
#endif

#define UART_BAUD_RATE	4800L*4 // check FUSY bits! poseble x2 on exit

void uart_init(void)
{

UCSRB = /* BV(RXCIE)| BV(TXCIE)|*/ (1<<RXCIE)|(1<<RXEN)|(1<<TXEN);

unsigned short bauddiv = ((F_CPU+(UART_BAUD_RATE*8L))/(UART_BAUD_RATE*16L)-1);
UBRRL = bauddiv;
UCSRA = (1<<U2X);

asm("nop");

/*
UCSRA = 0x00;
UCSRB = (1<<RXEN)|(1<<TXEN);	//Receive and transmit enable
*/
//UCSRC = (1<<URSEL)|(1<<UCSZ1)|(1<<UCSZ0);	

}

volatile unsigned char r_index=0, s_index=0;
//! UART Receive Complete Interrupt Handler
UART_INTERRUPT_HANDLER(SIG_UART_RECV)
{
	buffer[r_index] = UDR;
	if( ++r_index >= DATA_BLOCK_SIZE )
	{
		r_index = 0;
	}
}


void uart_transmit(unsigned char data)	//transmit byte by UART
{

while ( !( UCSRA & (1<<UDRE)) )
; 
UDR = data;
}

unsigned char uart_receive (void)		//receive byte by UART
{ 

  while ( !(UCSRA & (1<<RXC)) )
  ;
  return UDR;
}

void uart_transmit_message_P(PGM_P str)	//transmit string by UART
{
	while(pgm_read_byte(str) != 0) {
		uart_transmit(pgm_read_byte(str++));
		}
}

void uart_transmit_message(char* msg)	//transmit string by UART
{ unsigned char i;
  i=0;

  while ((i<128)&(msg[i]!=0x00) )
   {
    uart_transmit(msg[i]);
    i++;
   }
}

void spi_transmit (unsigned char data)	//transmit byte to SD|MMC card
{ unsigned char i;

 for (i=0;i<8;i++)
  {
   if ((data&0x80)==0x00) PORTB&=~_BV(DI);		//Set data bit
                     else PORTB|=_BV(DI);
   data=data<<1; 
   PORTB|=_BV(CLK);								//Strobe
   asm("nop"); 
   PORTB&=~_BV(CLK);
  }
}

unsigned char spi_receive (void)		//receive byte from SD|MMC card
{ unsigned char i, res=0;

 for (i=0;i<8;i++)
  {
   PORTB|=_BV(CLK);								//Start strobe
   res=res<<1;
   if ((PINB&_BV(DO))!=0x00) res=res|0x01;		//Get data bit
   PORTB&=~_BV(CLK);							//End strobe
   asm("nop");
  }
 return res;
}

unsigned char sd_cmd(char b0, char b1, char b2, char b3, char b4, char b5)	//Send command to SD|MMC card
{ unsigned char res;
  long int count;
 spi_transmit (b0);		//transmit index

 spi_transmit (b1);		//transmit argument
 spi_transmit (b2);
 spi_transmit (b3);
 spi_transmit (b4);

 spi_transmit (b5);		//transmit CRC
 count=0;
 do {					//wait to R1 responce
     res=spi_receive();
     count=count+1; 
    } while ( ((res&0x80)!=0x00)&&(count<0xffff) ); 

 return res;
}

unsigned char sd_card_init(void)
{ unsigned char i,temp;
  long int count;

  for (i=0;i<10;i++)							//80 strobes
   spi_transmit (0xff);   
	
  PORTB&=~_BV(CS);								//CS enable
	
  temp=sd_cmd (0x40,0x00,0x00,0x00,0x00,0x95);	//CMD0
  if (temp!=0x01) return 3;						//Exit, if bad responce
  spi_transmit (0xff); 

  count=0;
  do{
     temp=sd_cmd (0x41,0x00,0x00,0x00,0x00,0x95);	//CMD1
     spi_transmit (0xff); 
	 count=count+1;
  } while ( (temp!=0x00)&&(count<0xffff) );			//Wait 0x01 responce

 if (count>=0xffff) return 4;                
return 0;
}

unsigned char read_block (char* buff, unsigned long int addr)
{ 
  unsigned char temp, buff_idx;
  unsigned long int count;
  unsigned short int in_block_offs, j; 

  unsigned char a1 = addr >> 24;
  unsigned char a2 = addr >> 16;
  unsigned char a3 = addr >> 8 & 0xFE;	// 512 mask

  temp=sd_cmd (0x51,a1,a2,a3,0x0,0x95);		//Send CMD17
  if (temp!=0x00) return 5;					//Exit, if responce not 0x00
  spi_transmit (0xff); 
  count=0;
  do{										//Wait to data block starts
    temp=spi_receive(); 
    count++;
   } while ( (temp!=0xfe)&&(count<0xffff) );

  if (count>=0xffff) return 5;

  in_block_offs = addr % MMC_BLOCK_SIZE;
 
  for(j=0, buff_idx=0; j<MMC_BLOCK_SIZE; j++)
	{
		if( in_block_offs <= j && (in_block_offs + DATA_BLOCK_SIZE) > j )
		{
			buff[buff_idx] = spi_receive();
			buff_idx++;
		}
		else
			spi_receive();
	}

 spi_receive();								//receive CRC
 spi_receive();

 return 0;
}


#define FIRST_BLOCK_WRITE	1
#define LAST_BLOCK_WRITE	3
unsigned char write_block (char* buff, unsigned long int addr, char flg)
{
  unsigned char temp;
  unsigned char a1, a2, a3;
  long int count; 
  short int buff_idx; 
 if (FIRST_BLOCK_WRITE == flg){
  a1 = addr >> 24;
  a2 = addr >> 16;
  a3 = addr >> 8;

   temp=sd_cmd(0x58,a1,a2,a3,0x0,0x95);		//Send CMD24
   if (temp!=0x00) return 6;					//Exit, if responce non 0x00
   spi_transmit (0xff); 

   spi_transmit (0xfe);						//Start data block
 }

//  uart_transmit('\10');
// uart_transmit('\13');
//  uart_transmit(r_index+'0');
//  uart_transmit('\10');
//  uart_transmit('\13');
//  uart_transmit(s_index+'0');
  
  while( 1 )
	{
		if( r_index != s_index )
		{
			spi_transmit(buffer[s_index]);
			s_index++;
			if( s_index >= DATA_BLOCK_SIZE )
			{
				s_index = 0;
				break;
			}
		}
	}
 
  if (LAST_BLOCK_WRITE == flg){
    spi_transmit (0xff);						//Send CRC
    spi_transmit (0xff);
    temp=spi_receive();
    if ((temp&0x05)!=0x05) return 6;			//Exit, if data responce non 0x05

    count=0;
    do {										//Wait end of busy state responce
      temp=spi_receive();
      count++;
    }while ( (temp!=0xff)&&(count<0xffff) );

    if (count>=0xffff) return 6;
  }
 return 0;
}
  
/*  
  //
  //
  //
  int b_to_hex( unsigned char c, char *p)
  {
   unsigned char h = c >> 4;
   unsigned char l = c & 0x0F;
   
   if ( h > 9 )
       *p = h - 10 + 'A';
	else
       *p = h + '0';	   
	p++;
	
   if ( l > 9 )
       *p = l - 10 + 'A';
	else
       *p = l + '0';	   
	p++;
  
	*p = '\0';
	
  return 0;
  }

  //
  //
  //
  int b_to_dec( char c, char *p)
  {
    *p = (c/100) + '0';
	p++;
    *p = (c%100)/10 + '0';
	p++;
    *p = (c%10) + '0';
	p++;
	*p = '\0';
  return 0;
  }
  */
///
//	dont forgot would do enable interupt
//
int write_dict(unsigned int long first_block, unsigned long int cnt_blocks)
{
  unsigned char temp,ch;
  short int i;
  unsigned long int offs;

  for(offs=first_block*MMC_BLOCK_SIZE; offs<(MMC_BLOCK_SIZE*cnt_blocks+first_block*MMC_BLOCK_SIZE); offs+=MMC_BLOCK_SIZE)
  {
					// uart_transmit_message("write_block\r\n");

                  temp=write_block(buffer, offs, FIRST_BLOCK_WRITE);		//write buffer
				  if(temp)
				    {
					  ;//uart_transmit_message("write_block: CMD24 fail\r\n");
					  return 6;
					}

					
                  for (i=0;i<(MMC_BLOCK_SIZE/DATA_BLOCK_SIZE)-2L;i++)  // because FIRST_BLOCK_WRITE,LAST_BLOCK_WRITE
				  {
					temp=write_block(buffer, offs, 0);		//write buffer
                    if(temp==0x00) 
					  ;//uart_transmit_message("write_block: block writte succes\r\n");
				    else
					{				  
				      uart_transmit_message("write_block: CMD24 fail\r\n");
					  return 6;
					}
				  }
					
				  
					ch = buffer[0];
					temp=write_block(buffer, offs, LAST_BLOCK_WRITE);		//write buffer
                  if(temp==0x00) 
					;//uart_transmit_message("write_block: block writte succes\r\n");
				  else
					{				  
				      uart_transmit_message("write_block: CMD24 fail\r\n");
					  return 6;
					}
			uart_transmit(ch);
  }
  return 0;
}

#define SOCRAT_DICT_KEYWORD_BLOCK_OFFS		1L
#define SOCRAT_DICT_KEYWORD_BLOCK_SIZE		7042L
#define SOCRAT_DICT_ARTICLE_BLOCK_OFFS		7101L
#define SOCRAT_DICT_ARTICLE_BLOCK_SIZE		14084L


void main(void) __attribute__((__noreturn__));
void main(void)
{ 
  unsigned char temp;
  int i;

  
  PORTB=_BV(CS)|_BV(DO)|_BV(DI) /*|_BV(WP)|_BV(INS)*/; 	//Init ports
  DDRB=_BV(CS)|_BV(DI)|_BV(CLK);

  uart_init();			//Init UART
  sei();

 uart_transmit_message_P(PSTR("\n\n\n\nuart_init: ok\r\n"));
 temp=sd_card_init();	//Init card
 
 if (temp==0x00) 
	{
//			while(1){
//				char ch = 0;
//				ch = uart_receive();
//				uart_transmit(ch);
//			}
				//for (i=0;i<DATA_BLOCK_SIZE;i++)
				  //    buffer[i]=0x30+(i%10);		//Set buffer 12345...
/*
				temp = write_dict(SOCRAT_DICT_KEYWORD_BLOCK_OFFS, SOCRAT_DICT_KEYWORD_BLOCK_SIZE );	// by blocks in 512
                if(temp==0x00) 
				{
				    uart_transmit_message("succes:\r\n");
				}
				else 
				    uart_transmit_message("fail\r\n");
*/

				temp = write_dict(SOCRAT_DICT_ARTICLE_BLOCK_OFFS, SOCRAT_DICT_ARTICLE_BLOCK_SIZE );	// by blocks in 512
                if(temp==0x00) 
				{
				    uart_transmit_message_P(PSTR("succes:\r\n"));
				}
				else 
				    uart_transmit_message_P(PSTR("fail\r\n"));
				while (1) 
					;

}

 else if (temp==0x03) uart_transmit_message_P(PSTR("sd_card_init: CMD0 fail\r\n"));
 else uart_transmit_message_P(PSTR("sd_card_init: CMD1 fail\r\n"));

 while (1) 
  ;

// return 1; !!!
}
